from .editor import start_server, list_saved_files
import argparse
import sys
import os


def main():
    # Parse arguments
    parser = argparse.ArgumentParser()
    parser.add_argument("--editor",
                        help="starts Anubis editor",
                        action="store_true")
    parser.add_argument("-f", help="which folder to start")
    parser.add_argument("-p",
                        const=5000,
                        nargs="?",
                        help="on which port to start the webserver")
    parser.add_argument("--packages",
                        nargs='+',
                        help="libraries of boxes")
    args = parser.parse_args()

    # instantiate Anubis
    if args.editor:
        app = start_server(args)
        app.run(host="0.0.0.0", debug=True, port=args.p)
    else:
        from . import load_graph

        if args.f:
            filename = args.f
        else:
            filenames = list_saved_files()
            filename = "graphs/" + filenames[0] + ".json"
        print(f"Running graph: {filename}")

        ritual = load_graph(filename)
        ritual.run()



from .box import Function
from .slot import StrSlot, IntSlot, JsonSlot
import requests


# Arithmetic
class GetRequest(Function):
    name = "GET Request"
    inputs = [
        StrSlot("url"),
        JsonSlot("params")
    ]
    outputs = [
        StrSlot("text"),
        IntSlot("status code")
    ]

    def call(self, inputs):
        try:
            # if params are given
            if inputs[1] is not None:
                r = requests.get(inputs[0], params=inputs[1])
            else:
                r = requests.get(inputs[0])
            return r.text, r.status_code
        except requests.exceptions.ConnectionError:
            return "", 500



class Link:
    value_type = int
    name = "Link"
    gist = "Link"

    origin = None
    destination = None
    
    def __init__(self, origin, destination):
        import copy

        # get node, pin and value type
        if isinstance(origin, tuple):
            in_, in_pin, in_value_type = origin
        else:
            in_ = origin
            in_pin = 0
            in_value_type = origin.outputs[0].value_type
        # get node, pin and value type
        if isinstance(destination, tuple):
            out_, out_pin, out_value_type = destination
        else:
            out_ = destination
            out_pin = 0
            out_value_type = destination.inputs[0].value_type

        # check types are compatible (but not the RunSlot)
        if out_pin != 0 and out_value_type != "universal":
            try:
                if isinstance(in_value_type, tuple):
                    if isinstance(out_value_type, tuple):
                        assert any([in_type in out_value_type
                                    for in_type in in_value_type])
                    else:
                        assert out_value_type in in_value_type
                elif isinstance(out_value_type, tuple):
                    assert in_value_type in out_value_type
                else:
                    assert out_value_type == in_value_type
            except AssertionError:
                raise AssertionError(
                    f"Incompatible values between {in_.name} and {out_.name}")

        self.in_ = in_
        self.out_ = out_
        out_.inputs = copy.deepcopy(out_.inputs)
        out_.inputs[out_pin].origin = (in_.get_id(), in_pin)

    def to_json(self):
        return {
            "type": str(self.value_type),
            "name": self.name
        }



import pandas as pd


def convert_types(x):
    try:
        return x.__name__
    except AttributeError:
        return x.__str__()


class Slot:
    gist = "Slot"
    color = "#96F632"

    def __init__(self, name):
        self.name = name

    def to_json(self):
        return {
            "name": self.name,
            "types": [convert_types(i) for i in self.value_type],
            "color": self.color,
            "value": self.value
        }


class RunSlot(Slot):
    name = "Run"
    value_type = (bool,)
    value = True


class FloatSlot(Slot):
    value_type = (float, int)
    name = "Float"
    value = 0.


class IntSlot(Slot):
    value_type = (int,)
    name = "Int"
    value = 0


class AlphaNumSlot(Slot):
    value_type = (int, float, str)
    name = "AlphaNum"
    value = 0


class StrSlot(Slot):
    value_type = (str,)
    name = "Str"
    value = ""


class JsonSlot(Slot):
    value_type = (dict, list)
    name = "Json"
    value = "{}"


class ListSlot(Slot):
    value_type = (list,)
    name = "List"
    value = []


class ObjectSlot(Slot):
    value_type = (int, float, str, dict, list, bool, pd.DataFrame)
    name = "Object"
    value = {}


class UniversalSlot(Slot):
    value_type = "universal"
    name = "Universal"
    value = None


class DFSlot(Slot):
    value_type = (pd.DataFrame,)
    name = "DataFrame"
    value = None


class BoolSlot(Slot):
    value_type = (bool,)
    name = "Boolean"
    value = False



from .slot import (
    FloatSlot, StrSlot, JsonSlot, ObjectSlot, UniversalSlot,
    IntSlot, AlphaNumSlot, ListSlot, BoolSlot, RunSlot)
from enum import Enum
from collections import defaultdict
import json
import os


class State(Enum):
    IGNORE = 0


def check_inputs(func):
    def wrapper(instance, inputs=None):
        evaluate = True
        if inputs:
            for inp in inputs:
                if inp is State.IGNORE:
                    evaluate = False
                    break

        if evaluate:
            # start at index 1 because 0 is the RunSlot
            outputs = func(instance, inputs[1:])
            if not isinstance(outputs, tuple):
                outputs = (outputs,)
            assert len(outputs) >= len(instance.outputs)
        else:
            outputs = defaultdict(lambda: State.IGNORE)
        return outputs
    return wrapper


class Box:
    name = "Box"
    color = "#555555"
    inputs = []
    outputs = []
    parameters = []
    gist = "Box"
    description = ""

    def __init__(self, name=None, value=None):
        from copy import deepcopy

        if name is not None:
            assert isinstance(name, str)
            self.name = name

        if value is not None:
            assert isinstance(value, self.value_type)
            self.value = value

        self.primitive = self.__class__.__name__
        self.inputs = [RunSlot("run")] + self.inputs

    def to_json(self):
        res = {
            "name": self.name,
            "color": self.color,
            "inputs": [i.to_json() for i in self.inputs],
            "outputs": [i.to_json() for i in self.outputs],
            "parameters": [i.to_json() for i in self.parameters],
            "primitive": self.primitive,
            "gist": self.gist,
            "description": self.description
        }
        if self.gist == "Variable":
            res["value"] = self.value
        return res
    
    def from_json(self, data):
        pass
    
    def as_bool(self, value):
        return value in ["1", "true", "True", True, "yes", "Yes"]

    def _from(self, i):
        return (self, i, self.outputs[i].value_type)

    def _to(self, i):
        return (self, i, self.inputs[i].value_type)

    def get_id(self):
        return str(id(self))

    def __setattr__(self, key, value):
        if key == "value":
            if isinstance(self.value_type, tuple):
                super().__setattr__(key, self.value_type[0](value))
            else:
                super().__setattr__(key, self.value_type(value))
        else:
            super().__setattr__(key, value)

    @check_inputs
    def _call(self, *args, **kwargs):
        return self.call(*args, **kwargs)


class Function(Box):
    name = "Function"
    gist = "Function"
    color = "#49718B"
    inputs = []
    outputs = []


# Arithmetic
class Add(Function):
    name = "+"
    inputs = [
        FloatSlot("x"),
        FloatSlot("y")
    ]
    outputs = [
        FloatSlot("x + y")
    ]
    description = (
        """Add two floats"""
    )

    def call(self, inputs):
        return inputs[0] + inputs[1],


class Substract(Function):
    name = "-"
    inputs = [
        FloatSlot("x"),
        FloatSlot("y")
    ]
    outputs = [
        FloatSlot("x - y")
    ]
    description = (
        """Substract two floats"""
    )

    def call(self, inputs):
        return inputs[0] - inputs[1],


class GetItem(Function):
    name = "Get item"
    inputs = [
        JsonSlot("JSON")
    ]
    outputs = [
        FloatSlot("value"),
        FloatSlot("KeyError")
    ]
    parameters = [
        StrSlot("key")
    ]
    description = (
        """Return the value stored in the JSON at the `key` parameter."""
    )

    def call(self, inputs):
        if self.parameters[0] in inputs[0]:
            return inputs[0][self.parameters[0]], State.IGNORE
        return State.IGNORE, True


class SetItem(Function):
    name = "Set item"
    inputs = [
        JsonSlot("JSON"),
        ObjectSlot("value")
    ]
    outputs = [
        JsonSlot("JSON")
    ]
    parameters = [
        StrSlot("key")
    ]
    description = (
        """Store the value in the JSON at the `key` parameter."""
    )


    def call(self, inputs):
        inputs[0][self.parameters[0]] = inputs[1]
        return inputs[0]



# Variables
# ~~~~~~~~~

class Variable(Box):
    name = "Variable"
    gist = "Variable"
    color = "#598053"
    value = None


class Bool(Variable):
    name = "Bool"
    value = False
    outputs = [
        BoolSlot("Bool")
    ]
    description = (
        "Create a boolean variable"
    )
    def value_type(self, x):
        return self.as_bool(x)

    def call(self, inputs=None):
        return self.value


class Int(Variable):
    name = "Int"
    value_type = int
    value = 0
    outputs = [
        IntSlot("Int")
    ]
    description = (
        "Create an integer variable"
    )

    def call(self, inputs=None):
        return self.value


class Float(Variable):
    name = "Float"
    value_type = (float, int)
    value = 0
    outputs = [
        FloatSlot("Float")
    ]
    description = (
        "Create an floating point variable"
    )

    def call(self, inputs=None):
        return self.value


class List(Variable):
    name = "List"
    value_type = (list,)
    value = "[]"
    outputs = [
        ListSlot("List")
    ]
    description = (
        "Create a List variable"
    )

    def call(self, inputs=None):
        if isinstance(self.value, str):
            value = json.loads(self.value)
        elif isinstance(self.value, list):
            value = self.value
        return value


class Object(Variable):
    name = "Object"
    value_type = (float, int, str)
    value = 0
    outputs = [
        ObjectSlot("Object")
    ]
    description = (
        "Create an object variable"
    )

    def call(self, inputs=None):
        return self.value


class Str(Variable):
    name = "String"
    value_type = str
    value = ""
    outputs = [
        StrSlot("Str")
    ]
    description = (
        "Create a String variable"
    )

    def call(self, inputs=None):
        return self.value


class Json(Variable):
    name = "Json"
    value = {}
    outputs = [
        JsonSlot("Json")
    ]
    description = (
        "Create a JSON variable"
    )

    def value_type(self, x):
        if isinstance(x, str):
            return json.loads(x)
        elif isinstance(x, dict):
            return x

    def call(self, inputs=None):
        return self.value


class Environment(Variable):
    name = "Environment variable"
    value_type = str
    value = "PATH"
    outputs = [
        StrSlot("Str")
    ]
    description = (
        "Get an Environment variable by its name"
    )

    def call(self, inputs=None):
        return os.environ.get(self.value)


# Conditions
# ~~~~~~~~~~

class Condition(Box):
    name = "Event"
    gist = "Event"
    color = "#811012"
    inputs = []
    outputs = []
    parameters = []


class Branch(Condition):
    name = "Branch"
    inputs = [
        BoolSlot("condition")
    ]
    outputs = [
        BoolSlot("True"),
        BoolSlot("False"),
    ]
    description = (
        "Takes in a boolean, and execute the `True` branch if the boolean is True, "
        "the `False` branch is executed otherwise."
    )

    def call(self, inputs):
        if inputs[0]:
            return True, State.IGNORE
        else:
            return State.IGNORE, True


class GreaterThan(Condition):
    name = ">"
    inputs = [
        FloatSlot("x"),
        FloatSlot("y")
    ]
    outputs = [
        BoolSlot("x > y")
    ]
    description = (
        "Returns True if the first input is strictly greater than the second input."
    )

    def call(self, inputs):
        return inputs[0] > inputs[1]


class GreaterThanOrEqual(Condition):
    name = ">="
    inputs = [
        FloatSlot("x"),
        FloatSlot("y")
    ]
    outputs = [
        BoolSlot("x >= y")
    ]
    description = (
        "Returns True if the first input is greater than or equal to the second input."
    )

    def call(self, inputs):
        return inputs[0] >= inputs[1]


# Outputs
# ~~~~~~~

class Output(Box):
    name = "Output"
    value_type = str
    color = "#BA9303"
    inputs = [
        StrSlot("text")
    ]


class Print(Output):
    import pandas as pd

    name = "Print"
    value_type = lambda x: x
    inputs = [
        UniversalSlot("text")
    ]
    description = (
        "Print the given input in the console"
    )

    def call(self, inputs=None):
        if inputs is None:
            print(str(self.value))
        else:
            if isinstance(inputs[0], dict):
                from pprint import pprint
                pprint(inputs[0])
            else:
                print(str(inputs[0]))


class ToFile(Output):
    name = "Write to file"
    value_type = (str, int, float, dict, list)
    inputs = [
        ObjectSlot("content")
    ]
    parameters = [
        StrSlot("filename")
    ]
    description = (
        "Persist Python object to file. "
        "Can be any of string, integer, float, dict or list."
    )

    def call(self, inputs):
        content = inputs[0]
        if isinstance(content, dict):
            import json
            with open(self.parameters[0], "w") as f:
                json.dump(content, f)
        elif isinstance(content, list):
            with open(self.parameters[0], "w") as f:
                f.write("\n".join(content))
        else:
            with open(self.parameters[0], "w") as f:
                f.write(content)


# Casts
# ~~~~~

class Cast(Box):
    name = "Cast"
    gist = "Cast"
    color = "#111111"
    inputs = []
    outputs = []
    parameters = []


class ToInt(Cast):
    name = "Cast to Int"
    inputs = [
        AlphaNumSlot("Object")
    ]
    outputs = [
        IntSlot("Int")
    ]
    description = (
        "Cast any alpha numeric input to an integer"
    )

    def call(self, inputs=None):
        return int(inputs[0]),


class ToStr(Cast):
    name = "Cast to Str"
    inputs = [
        ObjectSlot("Object")
    ]
    outputs = [
        StrSlot("Str")
    ]
    description = (
        "Cast any python object to a string"
    )

    def call(self, inputs=None):
        return str(inputs[0]),


class ToList(Cast):
    name = "Cast to List"
    inputs = [
        ObjectSlot("Object")
    ]
    outputs = [
        ListSlot("List")
    ]
    description = (
        "Cast any python object to a list"
    )

    def call(self, inputs=None):
        return list([inputs[0]])


class ToJson(Cast):
    name = "Cast to JSON"
    inputs = [
        StrSlot("Object")
    ]
    outputs = [
        JsonSlot("JSON")
    ]
    description = (
        "Cast any python object to a JSON (Python Dict)"
    )

    def call(self, inputs=None):
        import json
        return json.loads(str(inputs[0])),


standard = [
    Bool,
    Float,
    Int,
    Str,
    Json,
    List,
    Environment,
    ToInt,
    ToStr,
    ToJson,
    ToFile,
    GetItem,
    SetItem,
    Branch,
    GreaterThan,
    GreaterThanOrEqual,
    Print,
    Add,
    Substract,
]


def load_package(packages):
    from .box import Function, Condition, Variable
    import inspect
    import importlib

    objs = []
    if packages:
        allowed_boxes = (
            Function.__class__,
            Condition.__class__,
            Variable.__class__)

        for package in packages:
            pkg = importlib.import_module(package)
            for name, obj in inspect.getmembers(pkg):
                if inspect.isclass(obj):
                    if obj.__module__ == package and isinstance(
                            obj, allowed_boxes):
                        objs.append(obj)
    return objs


def list_saved_files():
    from glob import glob
    filenames = glob("graphs/*.json")
    filenames = sorted(
        [item.split("/")[-1].split(".")[0] for item in filenames],
        reverse=True)
    return filenames

def start_server(args):
    from flask import Flask, send_from_directory, request, jsonify
    from . import Rite
    from datetime import datetime
    import logging
    import time
    import json
    import os

    # log = logging.getLogger('werkzeug')
    # log.setLevel(logging.ERROR)

    rite = Rite("project")
    rite += load_package(args.packages)

    # add imported boxes
    boxes = rite.to_json()

    # Start micro server
    app = Flask(__name__)

    @app.route("/")
    def index():
        return app.send_static_file("index.html")
    
    @app.route("/list", methods=["GET"])
    def list_saves():
        filenames = list_saved_files()
        return jsonify(filenames)

    @app.route("/load", methods=["GET"])
    def load():
        args = dict(request.args)
        if "id" not in args:
            return "ìd` parameter needed", 400

        file_id = args["id"]
        filename = os.path.join("graphs", file_id + ".json")
        with open(filename, "r") as f:
            data = json.load(f)
        return jsonify(data)

    @app.route("/save", methods=["POST"])
    def save():
        assert "nodes" in request.json
        assert "edges" in request.json

        nodes = request.json["nodes"]
        edges = request.json["edges"]

        if not os.path.exists("graphs"):
            os.makedirs("graphs")
        
        filename = datetime.now().strftime("%Y%m%d_%H%M%S") + ".json"
        filename = os.path.join("graphs/", filename)
        with open(filename, "w") as f:
            json.dump({
                "nodes": nodes,
                "edges": edges,
                "packages": args.packages
            }, f)
        
        print("File successfully saved")
        print("-----------------------")
        return "OK", 200

    @app.route("/run", methods=["POST"])
    def run():
        start_time = time.time()

        assert "nodes" in request.json
        assert "edges" in request.json

        nodes = request.json["nodes"]
        edges = request.json["edges"]

        ritual = rite.promote(nodes, edges)
        res = ritual.run()

        # show timing
        delta = time.time() - start_time
        print(f"Graph executed in {delta:.2f}s")
        print("-----------------------")

        return "OK", 200

    @app.route("/js/boxes.js")
    def send_box():
        return boxes


    @app.route("/js/<path:path>")
    def send_js(path):
        return send_from_directory("static/js", path)


    @app.route("/css/<path:path>")
    def send_css(path):
        return send_from_directory("static/css", path)


    # store variables
    if args.f is None:
        folder = "."
    else:
        folder = args.f

    return app



from .box import Function, Output, Cast
from .slot import StrSlot, DFSlot, BoolSlot, JsonSlot, ListSlot
import pandas as pd


class ReadDataFrame(Function):
    name = "Read DataFrame"
    inputs = [
        StrSlot("filename")
    ]
    outputs = [
        DFSlot("df")
    ]
    description = (
        "Read a DataFrame from disk: "
        "as pickle if filename extension is .p, "
        "as csv if filename extension is .csv, "
        "as hdf if filename extension is .hdf "
    )

    def call(self, inputs):
        filename = inputs[0]
        if ".p" == filename[-2:]:
            df = pd.read_pickle(filename)
        elif ".csv" == filename[-4:]:
            df = pd.read_csv(filename)
        elif ".hdf" == filename[-4:]:
            key = filename.split("/")[-1].split(".")[0]
            df = pd.read_hdf(filename, key)
        return df


class DFToCSV(Output):
    name = "Write DataFrame to CSV"
    inputs = [
        DFSlot("df")
    ]
    parameters = [
        StrSlot("filename"),
        BoolSlot("include index")
    ]

    def call(self, inputs):
        df = inputs[0]
        df.to_csv(self.parameters[0],
                  index=None)


class DFToJSON(Cast):
    name = "Cast DataFrame to JSON"
    inputs = [
        DFSlot("df")
    ]
    outputs = [
        JsonSlot("JSON")
    ]

    def call(self, inputs):
        df = inputs[0]
        return df.to_dict()


class DFToList(Cast):
    name = "Cast DataFrame to List"
    inputs = [
        DFSlot("df")
    ]
    outputs = [
        ListSlot("List")
    ]

    def call(self, inputs):
        df = inputs[0]
        return df.to_dict("records")



import setuptools

with open("README.md", "r") as f:
    long_description = f.read()

setuptools.setup(
    name="ritual",
    version="0.0.1",
    author="Maixent Chenebaux",
    author_email="max.chbx@gmail.com",
    description="Visual scripting tool for Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/kerighan/ritual",
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=["flask", "networkx", "pandas", "requests"],
    entry_points={
        "console_scripts": [
            "ritual=ritual.__main__:main"
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.5"
)



import networkx as nx


__version__ = "0.0.1"


class Rite:
    from .box import standard

    boxes = standard

    def __init__(self, name):
        assert isinstance(name, str)
        self.name = name

    def add_box(self, box):
        assert box.gist in ["Box", "Type", "Function", "Cast"]
        self.boxes.append(box)

    def __iadd__(self, boxes):
        if isinstance(boxes, list):
            for box in boxes:
                self.add_box(box)
        else:
            self.add_box(boxes)
        return self

    def to_json(self, filename=None):
        import json

        data = []
        for item in self.boxes:
            data.append(item().to_json())
        if filename is None:
            js = json.dumps(data)
            js = f"var boxes = {js};"
            return js
        else:
            with open(filename, "w") as f:
                js = json.dumps(data)
                js = f"var boxes = {js};"
                f.write(js)
    
    def build_name2box(self):
        self.name2box = {}
        for box in self.boxes:
            self.name2box[box.__name__] = box

    def __getitem__(self, key):
        return self.name2box[key]
    
    def promote(self, n, e):
        from .link import Link
        from copy import deepcopy

        self.build_name2box()
        ritual = Ritual()

        nodes = []
        for i in range(len(n)):
            name = n[i]["name"]
            primitive = n[i]["primitive"]
            node = deepcopy(self[primitive])(name)
            if "value" in n[i]:
                node.value = n[i]["value"]
            if "parameters" in n[i]:
                node.parameters = [j["value"] for j in n[i]["parameters"]]
            nodes.append(node)

        edges = []
        for edge in e:
            edges.append(Link(
                nodes[edge["outputNode"]]._from(edge["outputPin"]),
                nodes[edge["inputNode"]]._to(edge["inputPin"])
            ))
        ritual.create_graph(edges)
        return ritual


class Ritual():
    sorted_nodes = None
    verbose = True
    id2name = {}
    name2node = {}

    def create_graph(self, edges):
        if self.verbose:
            print("Creating graph...")

        di_edges = []
        for edge in edges:
            di_edges.append((edge.in_, edge.out_))

        self.G = nx.DiGraph()
        self.G.add_edges_from(di_edges)
        self.build_name2node()

        for node in self.G.nodes:
            self.id2name[node.get_id()] = node.name

        self.topsort()

    def build_name2node(self):
        self.name2node = {}
        for node in self.G.nodes:
            self.name2node[node.name] = node

    def feed(self, **kwargs):
        for key, value in kwargs.items():
            self.name2node[key].value = value

    def topsort(self):
        if self.verbose:
            print("Sorting nodes...")

        self.sorted_nodes = list(nx.topological_sort(self.G))
    
    def run(self, translate=True):
        assert self.sorted_nodes is not None

        if self.verbose:
            print("Running graph...")

        # call functions
        state = {}
        for i, node in enumerate(self.sorted_nodes):
            # get node id
            node_id = node.get_id()

            if len(node.inputs) == 0:
                if len(node.outputs) > 0:
                    state[node_id] = node._call()
                else:
                    node._call()
            else:
                args = [None] * len(node.inputs)
                for i in range(len(node.inputs)):
                    if hasattr(node.inputs[i], "origin"):
                        origin_id, pin = node.inputs[i].origin
                        args[i] = state[origin_id][pin]

                if len(node.outputs) > 0:
                    state[node_id] = node._call(args)
                else:
                    node._call(args)

        if not translate:
            return state
        
        state_translated = {}
        for key, value in state.items():
            state_translated[self.id2name[key]] = value
        return state_translated


def load_graph(filename):
    from .editor import load_package
    import json

    with open(filename, "r") as f:
        data = json.load(f)

    rite = Rite("project")
    rite += load_package(data["packages"])
    ritual = rite.promote(data["nodes"], data["edges"])
    return ritual
